from src.popoto.models.query import QueryException


class Field:
    type: type = str
    unique: bool = False
    indexed: bool = False
    value: str = None
    null: bool = False
    max_length: int = 256
    default: str = ""

    def __init__(self, **kwargs):
        full_kwargs = {  # default
            'type': str,
            'unique': True,
            'indexed': False,
            'value': None,
            'null': False,
            'max_length': 265,  # Redis limit is 512MB
            'default': None,
        }
        full_kwargs.update(kwargs)
        # if 'default' not in kwargs:
        #     full_kwargs['default'] = full_kwargs['type']()
        self.__dict__.update(full_kwargs)

    @classmethod
    def is_valid(cls, field, value) -> bool:
        if any([
            (value is None and not field.null),
            not isinstance(value, field.type),
            len(str(value)) > field.max_length,
        ]):
            return False
        return True

    @classmethod
    def pre_save(cls, model, field, value):
        pass

    @classmethod
    def post_save(cls, model, field, value):
        pass

    def get_filter_query_params(self, field_name: str) -> list:
        # todo: auto manage sets of db_keys to allow filter by any indexed field
        if self.indexed:
            return [
                f'{field_name}',
                f'{field_name}__isnull',
            ]
        return list()

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """
        raise QueryException("Sorry, query filter by any indexed field is not yet implemented.")




import msgpack
import logging
from src.popoto.redis_db import POPOTO_REDIS_DB

logger = logging.getLogger('POPOTO.Query')

class QueryException(Exception):
    pass


class Query:
    """
    an interface for db query operations using Popoto Models
    """
    model_class: 'Model'
    options: 'ModelOptions'

    def __init__(self, model_class: 'Model'):
        self.model_class = model_class
        self.options = model_class._meta

    def get(self, db_key=None, **kwargs):
        if db_key and '_auto_key' in self.options.key_field_names:
            raise QueryException(
                f"{self.model_class.__name__} does not define an explicit KeyField. Cannot perform query.get(key)"
            )

        if not db_key:
            for field_name, value in kwargs.items():
                if field_name not in self.options.indexed_field_names:
                    raise QueryException(
                        f"{field_name} is not an indexed field. Try using .filter({field_name}={value})"
                    )
            get_params = [
                (field_name, self.options.fields.get(field_name), value)
                for field_name, value in kwargs.items()
            ]

            db_key = ''
            # db_key = PopotoIndex.find(**get_params)

        instance = self.model_class(**{'db_key': db_key})
        if not instance.db_key:
            return None
        instance.load_from_db() or dict()
        if not len(instance._db_content):
            return None
        return instance

    def all(self):
        redis_db_keys_list = POPOTO_REDIS_DB.keys(f"{self.model_class.__name__}:*")
        return Query.get_many_objects(self.model_class, set(redis_db_keys_list))

    @classmethod
    def get_many_objects(cls, model: 'Model', db_keys: set):
        objects_list = []
        pipeline = POPOTO_REDIS_DB.pipeline()
        for db_key in db_keys:
            pipeline.hgetall(db_key)
        hashes_list = pipeline.execute()
        for redis_hash in hashes_list:
            objects_list.append(
                model(**{
                    key_b.decode("utf-8"): msgpack.unpackb(db_value_b)
                    for key_b, db_value_b in redis_hash.items()
                })
            )
        return objects_list

    def filter(self, **kwargs):
        """
           Access any and all filters for the fields on the model_class
           Run query using the given paramters
           return a list of model_class objects
        """
        from itertools import chain
        all_valid_filter_parameters = list(chain(*[
            field.get_filter_query_params(field_name)
            for field_name, field in self.options.fields.items()
        ]))

        db_keys_sets = []

        for field_name, field in self.options.fields.items():
            from src.popoto import KeyField, GeoField, SortedField
            if isinstance(field, KeyField):

                logger.debug(f"query on {field_name}")

                # intersection of field params and filter kwargs
                params_for_keyfield = set(kwargs.keys()) & set(field.get_filter_query_params(field_name))

                logger.debug({k: kwargs[k] for k in params_for_keyfield})

                key_set = KeyField.filter_query(
                    self.model_class, field_name, **{k: kwargs[k] for k in params_for_keyfield}
                )
                if len(key_set):
                    db_keys_sets.append(key_set)
        logger.debug(db_keys_sets)
        if not db_keys_sets:
            return []
        return Query.get_many_objects(self.model_class, set.intersection(*db_keys_sets))



from datetime import datetime

import msgpack
import redis

import logging

from .query import Query
from ..fields.key_field import KeyField
from ..fields.field import Field
from ..fields.sorted_field import SortedField
from ..redis_db import POPOTO_REDIS_DB, ENCODING

logger = logging.getLogger('POPOTO.model_base')


class ModelException(Exception):
    pass


class ModelOptions:
    def __init__(self, model_name):
        self.model_name = model_name
        self.hidden_fields = {}
        self.explicit_fields = {}
        self.key_field_names = []
        self.list_field_names = []
        # self.set_field_names = []
        self.sorted_field_names = []
        self.geo_field_names = []
        self.indexed_field_names = []

        # todo: should this be a dict of related objects or just a list of field names?
        # self.related_fields = {}  # model becomes graph node

        # todo: allow customizing this in model.Meta class
        self.db_class_key = self.model_name

        self.abstract = False
        self.unique_together = []
        self.index_together = []
        self.parents = []
        self.auto_created = False
        self.base_meta = None

    def add_field(self, field_name: str, field: Field):
        if field_name.startswith("_") and field_name not in self.hidden_fields:
            self.hidden_fields[field_name] = field
        elif field_name not in self.explicit_fields:
            self.explicit_fields[field_name] = field
        else:
            raise ModelException(f"{field_name} is already a Field on the model")

        if field.indexed:
            self.indexed_field_names.append(field_name)

        if isinstance(field, KeyField):
            self.key_field_names.append(field_name)
        elif isinstance(field, SortedField):
            self.sorted_field_names.append(field_name)
        # elif isinstance(field, ListField):
        #     self.list_field_names.append(field_name)
        # elif isinstance(field, GeoField):
        #     self.geo_field_names.append(field_name)

    @property
    def fields(self) -> dict:
        return {**self.explicit_fields, **self.hidden_fields}

    @property
    def field_names(self) -> list:
        return list(self.fields.keys())

    @property
    def db_key_length(self):
        return 1 + len(self.key_field_names)

    def get_db_key_position(self, field_name):
        return 1 + self.key_field_names.index(field_name)


class ModelBase(type):
    """Metaclass for all Redis models."""

    def __new__(cls, name, bases, attrs, **kwargs):

        # Initialization is only performed for a Model and its subclasses
        parents = [b for b in bases if isinstance(b, ModelBase)]
        if not parents:
            return super().__new__(cls, name, bases, attrs)

        # logger.debug({k: v for k, v in attrs.items() if not k.startswith('__')})
        module = attrs.pop('__module__')
        new_attrs = {'__module__': module}
        attr_meta = attrs.pop('Meta', None)
        options = ModelOptions(name)
        options.parents = parents

        for obj_name, obj in attrs.items():
            if obj_name.startswith("__"):
                # builtin or inherited private vars and methods
                new_attrs[obj_name] = obj

            elif isinstance(obj, Field):
                # save field instance
                # attr will be overwritten as a field.type
                # model will handle this and set default values
                options.add_field(obj_name, obj)

            elif callable(obj) or hasattr(obj, '__func__') or hasattr(obj, '__set__'):
                # a callable method or property
                new_attrs[obj_name] = obj

            elif obj_name.startswith("_"):
                # a private static attr not to be saved in the db
                new_attrs[obj_name] = obj

            else:
                raise ModelException(
                    f"public model attributes must inherit from class Field. "
                    f"Try using a private var (eg. _{obj_name})_"
                )

        # todo: handle multiple inheritance
        # for base in parents:
        #     for field_name, field in base.auto_fields.items():
        #         options.add_field(field_name, field)

        new_class = super().__new__(cls, name, bases, new_attrs)

        options.abstract = getattr(attr_meta, 'abstract', False)
        options.meta = attr_meta or getattr(new_class, 'Meta', None)
        options.base_meta = getattr(new_class, '_meta', None)
        new_class._meta = options
        setattr(new_class, 'query', Query(new_class))
        setattr(new_class, 'objects', new_class.query)
        return new_class


class Model(metaclass=ModelBase):
    query: Query

    def __init__(self, **kwargs):
        cls = self.__class__
        # self._ttl = kwargs.get('ttl', None)
        # self._expire_at = kwargs.get('expire_at', None)

        # allow init kwargs to set any base parameters
        self.__dict__.update(kwargs)

        # add auto KeyField if needed
        if not len(self._meta.key_field_names):
            self._meta.add_field('_auto_key', KeyField(auto=True, key=cls.__name__))

        # set defaults
        for field_name, field in self._meta.fields.items():
            setattr(self, field_name, field.default)

        # set field values from init kwargs
        for field_name in self._meta.fields.keys() & kwargs.keys():
            setattr(self, field_name, kwargs.get(field_name))

        # todo: handle some key management?
        # perhaps need to prepare to handle events such as key change, partial key search, ...

        self._ttl = None  # todo: set default in child Meta class
        self._expire_at = None  # todo: datetime? or timestamp?

        # validate initial attributes
        self.validate(null_check=False)  # exclude null, will validate null values on pre-save
        self._db_content = dict()  # empty until self.load_from_db() or self.save() called

    @property
    def db_key(self):
        """
        the db key must include the class name - equivalent to db table name
        like it or not, keys append alphabetically.
        if needed, can propose feature to include order=int param on KeyField to force an order
        """
        return f"{self._meta.db_class_key}:" + ":".join([
            getattr(self, key_field_name) for key_field_name in sorted(self._meta.key_field_names)
        ])

    def __str__(self):
        return str(self.db_key)

    def __eq__(self, other):
        return repr(self) == repr(other)

    # @property
    # def field_names(self):
    #     return [
    #         k for k, v in self.__dict__.items()
    #         if all([not k.startswith("_"), k + "_meta" in self.__dict__])
    #     ]

    def validate(self, null_check=True) -> bool:
        """
            todo: validate values
            - field.type ✅
            - field.null ✅
            - field.max_length ✅
            - ttl, expire_at - todo
        """

        for field_name in self._meta.field_names:
            # type check the field values against their class specified type, unless null/None

            if all([
                getattr(self, field_name) is not None,
                not isinstance(getattr(self, field_name), self._meta.fields[field_name].type)
            ]):
                try:
                    setattr(self, field_name, self._meta.fields[field_name].type(getattr(self, field_name)))
                    if not isinstance(getattr(self, field_name), self._meta.fields[field_name].type):
                        raise TypeError(f"{field_name} is not type {self._meta.fields[field_name].type}. ")
                except TypeError as e:
                    logger.error(
                        f"{str(e)} \n Change the value or modify type on {self.__class__.__name__}.{field_name}"
                    )
                    return False

            # check non-nullable fields
            if null_check and \
                    self._meta.fields[field_name].null is False and \
                    getattr(self, field_name) is None:
                error = f"{field_name} is None/null. Set a value or set null=True on {self.__class__.__name__}.{field_name}"
                logger.error(error)
                return False

            # validate str max_length
            if self._meta.fields[field_name].type == str and \
                    getattr(self, field_name) and \
                    len(getattr(self, field_name)) > self._meta.fields[field_name].max_length:
                error = f"{field_name} is greater than max_length={self._meta.fields[field_name].max_length}"
                logger.error(error)
                return False

            if self._ttl and self._expire_at:
                raise ModelException("Can set either ttl and expire_at. Not both.")
        return True

    def save(self, pipeline: redis.client.Pipeline = None, ttl=None, expire_at=None, ignore_errors: bool = False, *args,
             **kwargs):
        """
            Default save method. Uses Redis HSET command with key, dict of values, ttl.
        """
        if not self.validate():
            error_message = "Model instance parameters invalid. Failed to save."
            if ignore_errors:
                logger.error(error_message)
            else:
                raise ModelException(error_message)
            return pipeline or 0

        ttl, expire_at = (ttl or self._ttl), (expire_at or self._expire_at)
        hset_mapping = {
            str(k).encode(ENCODING): msgpack.packb(v)
            for k, v in self.__dict__.items() if k in self._meta.fields
        }
        self._db_content = hset_mapping
        if isinstance(pipeline, redis.client.Pipeline):
            pipeline = pipeline.hset(self.db_key, mapping=hset_mapping)
            pipeline = pipeline if ttl is None else pipeline.expire(self.db_key, ttl)
            pipeline = pipeline if expire_at is None else pipeline.expire_at(self.db_key, expire_at)
            return pipeline
        else:
            db_response = POPOTO_REDIS_DB.hset(self.db_key, mapping=hset_mapping)
            if ttl is not None: POPOTO_REDIS_DB.expire(self.db_key, ttl)
            if expire_at is not None: POPOTO_REDIS_DB.expireat(self.db_key, ttl)
            return db_response

    @classmethod
    def create(cls, *args, **kwargs):
        instance = cls(*args, **kwargs)
        instance.save()
        return instance

    @classmethod
    def get(cls, db_key: str = None, **kwargs):
        return cls.query.get(db_key=db_key, **kwargs)

    def __repr__(self):
        return str({k: v for k, v in self.__dict__.items() if k in self._meta.fields})

    def load_from_db(self):
        self._db_content = POPOTO_REDIS_DB.hgetall(self.db_key)
        for key_b, db_value_b in self._db_content.items():
            setattr(self, key_b.decode("utf-8"), msgpack.unpackb(db_value_b))

    def revert(self):
        self.load_from_db()

    def delete(self, pipeline=None, *args, **kwargs):
        self._db_content = dict()
        if pipeline is not None:
            pipeline = pipeline.delete(self.db_key)
            return pipeline
        else:
            db_response = POPOTO_REDIS_DB.delete(self.db_key)
            if db_response >= 0:
                return True



from time import time
from .field import Field
import uuid

from collections import namedtuple

from ..models.query import QueryException
from ..redis_db import POPOTO_REDIS_DB

Coordinates = namedtuple('Coordinates', 'latitude longitude')

class GeoField(Field):
    """
    A field that enables geospatial data and geospatial search
    required: latitude, longitude
    """
    type: type = tuple
    latitude: float = None
    longitude: float = None
    default: tuple = (None, None)
    null: bool = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        new_kwargs = {  # default
            'latitude': None,
            'longitude': None,
            'null': True,
        }
        new_kwargs.update(kwargs)
        for k in new_kwargs:
            setattr(self, k, new_kwargs[k])

    def get_filter_query_params(self, field_name):
        return super().get_filter_query_params(field_name) + [
            f'{field_name}_latitude',
            f'{field_name}_longitude',
            f'{field_name}_coordinates',
            f'{field_name}_radius',
            f'{field_name}_radius_units',
        ]

    @classmethod
    def filter_query(cls, model: 'Model', field_name: str, **query_params) -> set:
        """
        :param model: the popoto.Model to query from
        :param field_name: the name of the field being filtered on
        :param query_params: dict of filter args and values
        :return: set{db_key, db_key, ..}
        """

        coordinates = Coordinates(None, None)
        member, radius, units = None, 0, 'm'
        for query_param, query_value in query_params.items():

            if '_coordinates' in query_param:
                if not isinstance(query_value, tuple) or not len(query_value) == 2:
                    raise QueryException(f"{query_param} must be assigned a tuple = (latitude, longitude)")
                coordinates = Coordinates(query_value[0], query_value[1])

            elif '_latitude' in query_param:
                coordinates.latitude = query_value
            elif '_longitude' in query_param:
                coordinates.longitude = query_value

            elif '_member' in query_param:
                if not isinstance(query_value, model):
                    raise QueryException(f"{query_param} must be assigned a tuple = (latitude, longitude)")
                member = query_value

            elif query_param == 'radius':
                radius = query_value

            elif query_param == 'units':
                if query_value not in ['m', 'km', 'ft', 'mi']:
                    raise QueryException(f"{query_param} must be one of m|km|ft|mi ")
                units = query_value

        if member:
            redis_db_keys_list = POPOTO_REDIS_DB.georadiusbymember(
                model._meta.db_class_key, member=member.db_key,
                radius=radius, units=units
            )

        elif bool(coordinates.latitude and coordinates.longitude):
            redis_db_keys_list = POPOTO_REDIS_DB.georadius(
                model._meta.db_class_key,
                longitude=coordinates.longitude, latitude=coordinates.latitude,
                radius=radius, units=units
            )
        else:
            raise QueryException(f"missing one or more required parameters. "
                                 f"geofilter requires either coordinates or instance of the same model")

        return set(redis_db_keys_list)
